//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "MandelThread.h"
#include <math.h>
#pragma package(smart_init)
//---------------------------------------------------------------------------

//   Important: Methods and properties of objects in VCL can only be
//   used in a method called using Synchronize, for example:
//
//      Synchronize(UpdateCaption);
//
//   where UpdateCaption could look like:
//
//      void __fastcall TMandelThread::UpdateCaption()
//      {
//        Form1->Caption = "Updated in a thread";
//      }
//---------------------------------------------------------------------------


__fastcall TMandelThread::TMandelThread(bool CreateSuspended)
  : TThread(CreateSuspended)
{
  msg_queue = new CRegistry(20);
  Reset();
}

__fastcall TMandelThread::~TMandelThread()
{
  delete msg_queue;
}
//---------------------------------------------------------------------------
void __fastcall TMandelThread::Execute()
{
  do
  {
    int msg;

    if (!msg_queue->is_empty())
    {
      msg = msg_queue->get();
      if (msg>0)
      {
         start = false;
         in_progress = true;

         start_time = GetTickCount();
         switch (msg)
         {
            case MANDEL2: Mandel2();
              break;
            case MANDEL3: Mandel3();
              break;
            case MANDEL4: Mandel4();
              break;
            case JULIA2: Julia2();
              break;
            case JULIA3: Julia3();
              break;
            case JULIA4: Julia4();
              break;
         }
         time = GetTickCount()-start_time;
         msg_queue->put(STOP);
      }
      if (msg==STOP) in_progress=false;
    }
    Sleep(50);
  } while (!Terminated);
}
//---------------------------------------------------------------------------

void TMandelThread::Reset()
{
  xmin = -1.5;
  xmax = 1.5;
  ymin = 1.5;
  ymax = -1.5;
  iter = 255;
 // pBMP = NULL;
  poc_vyp = 0;
  in_progress = false;
  progress = 0;
  msg_queue->clear();
  Set_iter(255);
  time = 0;

  //if (output) delete output;
  //output = NULL;
  //outputx=0;
  //outputy=0;
}

void TMandelThread::Start(int exp, int x, int y)
{
  data.setDimensions(x,y);
  msg_queue->put(exp);
  in_progress = false;
}
/*
void TMandelThread::Set_pBMP(Graphics::TBitmap *p)
{
  pBMP = p;
  data.setDimensions(p->Width,p->Height);
} */

void TMandelThread::Set_xmin(double x)
{
  msg_queue->put(STOP);
  in_progress = false;
  xmin = x;
}

void TMandelThread::Set_xmax(double x)
{
  msg_queue->put(STOP);
  in_progress = false;
  xmax = x;
}

void TMandelThread::Set_ymin(double x)
{
  msg_queue->put(STOP);
  in_progress = false;
  ymin = x;
}

void TMandelThread::Set_ymax(double x)
{
  msg_queue->put(STOP);
  in_progress = false;
  ymax = x;
}

double TMandelThread::Get_xmin()
{
  return xmin;
}

double TMandelThread::Get_xmax()
{
  return xmax;
}

double TMandelThread::Get_ymin()
{
  return ymin;
}

double TMandelThread::Get_ymax()
{
  return ymax;
}

unsigned int TMandelThread::Get_iter()
{
  return iter;
}

void TMandelThread::Set_iter(unsigned int i)
{
  msg_queue->put(STOP);
  in_progress = false;
  iter = i;
  if (re) delete re;
  if (im) delete im;
  if (abs) delete abs;
  im = new double[iter];
  re = new double[iter];
  abs= new double[iter];
}

/*Graphics::TBitmap * TMandelThread::Get_pBMP()
{
  return pBMP;
} */

double TMandelThread::Get_progress()
{
  return progress;
}

bool __fastcall TMandelThread::In_Progress()
{
  return in_progress;
}

void TMandelThread::IterSeq(int type,int x, int y, double r)
{
  double x1 = XtoRe(x);
  double y1 = YtoIm(y);

  double p;
  double q;

  re[0] = x1;
  im[0] = y1;
  abs[0] = sqrt(x1*x1+y1*y1);

  switch (type)
  {
    case MANDEL2:
    case MANDEL3:
    case MANDEL4:
      p = x1;
      q = y1;
    break;

    case JULIA2:
    case JULIA3:
    case JULIA4:
      p = cr;
      q = ci;
    break;
  }


  int k=1;

  double x2,y2;

  while (k<iter)
  {
    switch (type)
    {
      case MANDEL2:
      case JULIA2:
        x2 = x1*x1-y1*y1  +p;
        y2 = 2*x1*y1      +q;
        break;
      case MANDEL3:
      case JULIA3:
        x2 = x1*x1*x1-3*x1*y1*y1+p;
        y2 = 3*x1*x1*y1-y1*y1*y1+q;
        break;
      case MANDEL4:
      case JULIA4:
        x2 = x1*x1*x1*x1-6*x1*x1*y1*y1+y1*y1*y1*y1 + p;
        y2 = 4*x1*x1*x1*y1-4*x1*y1*y1*y1 + q;
        break;
    }

    x1=x2;
    y1=y2;

    re[k]= x1;
    im[k]= y1;
    abs[k] = sqrt(x1*x1+y1*y1);

    if (abs[k]>r) break;

    k++;
  }

  while (k<iter)
  {
    re[k]= re[k-1];
    im[k]= im[k-1];
    abs[k] = abs[k-1];
    k++;
  }
}

double * __fastcall TMandelThread::Get_pRe()
{
  return re;
}

double * __fastcall TMandelThread::Get_pIm()
{
  return im;
}

double * __fastcall TMandelThread::Get_pAbs()
{
  return abs;
}

void TMandelThread::ZoomRect(int x1, int y1,int x2, int y2)
{
  msg_queue->put(STOP);
  in_progress = false;
  
  if (x1==x2 || y1==y2) return;
  if (x1>x2)
  {
    int a = x1;
    x1 = x2;
    x2 = a;
  }
  if (y1>y2)
  {
    int a = y1;
    y1 = y2;
    y2 = a;
  }

  //double dx = (xmax - xmin)/data.getX();
  //double dy = (ymax - ymin)/data.getY();
  double xminnew = XtoRe(x1);
  double yminnew = YtoIm(y1);
  double xmaxnew = XtoRe(x2);
  double ymaxnew = YtoIm(y2);

  xmin = xminnew;
  xmax = xmaxnew;
  ymin = yminnew;
  ymax = ymaxnew;
}

unsigned TMandelThread::GetLastCalcTime()
{
  return time;
}

void TMandelThread::ZoomPct(double zoom)
{                        //zoom=1 => no zoom
  double xminnew = (xmin + xmax)/2 - (xmax-xmin)/(2*zoom);
  double xmaxnew = (xmin + xmax)/2 + (xmax-xmin)/(2*zoom);

  double yminnew = (ymin + ymax)/2 - (ymax-ymin)/(2*zoom);
  double ymaxnew = (ymin + ymax)/2 + (ymax-ymin)/(2*zoom);

  xmin = xminnew;
  xmax = xmaxnew;
  ymin = yminnew;
  ymax = ymaxnew;
}

int TMandelThread::Get_PocVyp(void)
{
  return poc_vyp;
}

double TMandelThread::XtoRe(int x)
{
  double dx = (xmax - xmin)/data.getX();
  return xmin+x*dx;
}

double TMandelThread::YtoIm(int y)
{
  double dy = (ymax - ymin)/data.getY();
  return (ymin+y*dy);
}

void TMandelThread::setCr(double r)
{
  cr=r;
}

void TMandelThread::setCi(double i)
{
  ci=i;
}


bool TMandelThread::Mandel2()
{
  unsigned nx,ny;       //position in bitmap
  double dx,dy,p,q;
  double x1,y1,r;
  unsigned *C;   //pointer to a line in bitmap
  int k;
  unsigned x,y;    //bitmap dimensions
  double x2;
  double y2;
  double max_r;        //max polomer, pro moznou konverg.

   CDataItem* item=NULL;

   x=data.getX();
   y=data.getY();

   dx = (xmax - xmin)/x;
   dy = (ymax - ymin)/y;

   poc_vyp=0;           //pocet vypoctu
   item =data.getDataPtr();
   for(ny=0;ny<y;ny++)
   {
      //C=(unsigned*)pBMP->ScanLine[ny];
      for(nx=0;nx<x;nx++)
      {
         k=0;

         x1 = xmin+nx*dx;
         y1 = ymin+ny*dy;

         p =x1;
         q =y1;

         r=p*p+q*q;
         if (r>4) max_r=r;
         else   max_r=4;

         r=x1*x1+y1*y1;

         while ((r<=max_r) && (k<iter))
         {
             x2 = x1*x1-y1*y1+p;
             y2 = 2*x1*y1+q;

             x1=x2;
             y1=y2;

             k++;

             r = x1*x1+y1*y1;
             poc_vyp++;
         }

         if (!in_progress) return false;

         if (k==iter) k=0;

         item->numiter = k;
         item++;
      }
      if (y>0) progress = 100*ny/((double)y-1);
      data.setLinesFilled(ny);
   }
   return true;
}

bool TMandelThread::Julia2()
{
  unsigned nx,ny;       //position in bitmap
  double dx,dy;
  double x1,y1,r;
  int k;
  unsigned x,y;    //bitmap dimensions
  double x2;
  double y2;
  double max_r;        //max polomer, pro moznou konverg.

   CDataItem* item=NULL;

   x=data.getX();
   y=data.getY();

   dx = (xmax - xmin)/x;
   dy = (ymax - ymin)/y;

   poc_vyp=0;           //pocet vypoctu
   item =data.getDataPtr();

   r=cr*cr+ci*ci;
   if (r>4) max_r=r;
   else   max_r=4;

   for(ny=0;ny<y;ny++)
   {
      for(nx=0;nx<x;nx++)
      {
         k=0;

         x1 = xmin+nx*dx;
         y1 = ymin+ny*dy;

         r=x1*x1+y1*y1;

         while ((r<=max_r) && (k<iter))
         {
             x2 = x1*x1-y1*y1+cr;
             y2 = 2*x1*y1+ci;

             x1=x2;
             y1=y2;

             k++;

             r = x1*x1+y1*y1;
             poc_vyp++;
         }

         if (!in_progress) return false;

         if (k==iter) k=0;

         item->numiter = k;
         item++;
      }
      if (y>0) progress = 100*ny/((double)y-1);
      data.setLinesFilled(ny);
   }
   return true;
}

bool TMandelThread::Mandel3()
{
  unsigned nx,ny;       //position in bitmap
  double dx,dy,p,q;
  double x1,y1,r;
  //unsigned *C;   //pointer to a line in bitmap
  int k;
  unsigned x,y;    //bitmap dimensions
  double x2;
  double y2;
  double max_r;        //max polomer, pro moznou konverg.

  CDataItem* item=NULL;

   x=data.getX();
   y=data.getY();

   dx = (xmax - xmin)/x;
   dy = (ymax - ymin)/y;

   poc_vyp=0;           //pocet vypoctu
   item =data.getDataPtr();
   for(ny=0;ny<y;ny++)
   {
      for(nx=0;nx<x;nx++)
      {
         k=0;

         x1 = xmin+nx*dx;
         y1 = ymin+ny*dy;

         p =x1;
         q =y1;

         r=p*p+q*q;
         if (r>4) max_r=r;
         else   max_r=4;

         r=x1*x1+y1*y1;

         while ((r<=max_r) && (k<iter))
         {
             x2 = x1*x1*x1-3*x1*y1*y1+p;
             y2 = 3*x1*x1*y1-y1*y1*y1+q;

             x1=x2;
             y1=y2;

             k++;

             r = x1*x1+y1*y1;
             poc_vyp++;
         }

         if (!in_progress) return false;

         if (k==iter) k=0;
         item->numiter = k;
         item++;
      }
      if (y>0)
         progress = 100*ny/((double)y-1);
      data.setLinesFilled(ny);
   }
   return true;
}


bool TMandelThread::Julia3()
{
  unsigned nx,ny;       //position in bitmap
  double dx,dy;
  double x1,y1,r;
  int k;
  unsigned x,y;    //bitmap dimensions
  double x2;
  double y2;
  double max_r;        //max polomer, pro moznou konverg.

   CDataItem* item=NULL;

   x=data.getX();
   y=data.getY();

   dx = (xmax - xmin)/x;
   dy = (ymax - ymin)/y;

   poc_vyp=0;           //pocet vypoctu
   item =data.getDataPtr();

   r=cr*cr+ci*ci;
   if (r>4) max_r=r;
   else   max_r=4;

   for(ny=0;ny<y;ny++)
   {
      for(nx=0;nx<x;nx++)
      {
         k=0;

         x1 = xmin+nx*dx;
         y1 = ymin+ny*dy;

         r=x1*x1+y1*y1;

         while ((r<=max_r) && (k<iter))
         {
             x2 = x1*x1*x1-3*x1*y1*y1+cr;
             y2 = 3*x1*x1*y1-y1*y1*y1+ci;

             x1=x2;
             y1=y2;

             k++;

             r = x1*x1+y1*y1;
             poc_vyp++;
         }

         if (!in_progress) return false;

         if (k==iter) k=0;

         item->numiter = k;
         item++;
      }
      if (y>0) progress = 100*ny/((double)y-1);
      data.setLinesFilled(ny);
   }
   return true;
}

bool TMandelThread::Mandel4()
{
  unsigned nx,ny;       //position in bitmap
  double dx,dy,p,q;
  double x1,y1,r;
  //unsigned *C;   //pointer to a line in bitmap
  int k;
  unsigned x,y;    //bitmap dimensions
  double x2;
  double y2;
  double max_r;        //max polomer, pro moznou konverg.

  CDataItem* item=NULL;

   x=data.getX();
   y=data.getY();

   dx = (xmax - xmin)/x;
   dy = (ymax - ymin)/y;

   poc_vyp=0;           //pocet vypoctu
   item =data.getDataPtr();
   for(ny=0;ny<y;ny++)
   {
      for(nx=0;nx<x;nx++)
      {
         k=0;

         x1 = xmin+nx*dx;
         y1 = ymin+ny*dy;

         p =x1;
         q =y1;

         r=p*p+q*q;
         if (r>4) max_r=r;
         else   max_r=4;

         r=x1*x1+y1*y1;

         while ((r<=max_r) && (k<iter))
         {
             x2 = x1*x1*x1*x1 - 6*x1*x1*y1*y1 + y1*y1*y1*y1 + p;
             y2 = 4*x1*x1*x1*y1-4*x1*y1*y1*y1 + q;

             x1=x2;
             y1=y2;

             k++;

             r = x1*x1+y1*y1;
             poc_vyp++;
         }

         if (!in_progress) return false;

         if (k==iter) k=0;
         item->numiter = k;
         item++;
      }
      if (y>0)
         progress = 100*ny/((double)y-1);
      data.setLinesFilled(ny);
   }
   return true;
}

bool TMandelThread::Julia4()
{
  unsigned nx,ny;       //position in bitmap
  double dx,dy;
  double x1,y1,r;
  int k;
  unsigned x,y;    //bitmap dimensions
  double x2;
  double y2;
  double max_r;        //max polomer, pro moznou konverg.

   CDataItem* item=NULL;

   x=data.getX();
   y=data.getY();

   dx = (xmax - xmin)/x;
   dy = (ymax - ymin)/y;

   poc_vyp=0;           //pocet vypoctu
   item =data.getDataPtr();

   r=cr*cr+ci*ci;
   if (r>4) max_r=r;
   else   max_r=4;

   for(ny=0;ny<y;ny++)
   {
      for(nx=0;nx<x;nx++)
      {
         k=0;

         x1 = xmin+nx*dx;
         y1 = ymin+ny*dy;

         r=x1*x1+y1*y1;

         while ((r<=max_r) && (k<iter))
         {
             x2 = x1*x1*x1*x1-6*x1*x1*y1*y1+y1*y1*y1*y1 + cr;
             y2 = 4*x1*x1*x1*y1-4*x1*y1*y1*y1 + ci;

             x1=x2;
             y1=y2;

             k++;

             r = x1*x1+y1*y1;
             poc_vyp++;
         }

         if (!in_progress) return false;

         if (k==iter) k=0;

         item->numiter = k;
         item++;
      }
      if (y>0) progress = 100*ny/((double)y-1);
      data.setLinesFilled(ny);
   }
   return true;
}

